/* CommandLine Send Mail
 * Copyright (C) 2008-2009 Michael Stum
 * http://www.Stum.de
 * http://cmdtools.codeplex.com
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation, Version 2, June 1991.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, you can find it on the GNU Website:
 * http://www.gnu.org/licenses/old-licenses/gpl-2.0.txt
 */

using System;
using System.Collections.Generic;
using System.Text;
using System.Net.Mail;
using System.Xml;
using System.IO;
using System.Net;
using System.Reflection;
using System.Diagnostics;
using System.Globalization;

namespace stumde.cmdtools.cmdsendmail
{
    /// <summary>
    /// The cmdsendmail Console Application
    /// </summary>
    class Program
    {
        /// <summary>
        /// The Xml from the command line file
        /// </summary>
        static XmlDocument _settingsXml;

        /// <summary>
        /// Be Verbose or not
        /// </summary>
        static bool _verbose;

        /// <summary>
        /// Displays "About/Copyright" information
        /// </summary>
        static void DisplayAbout()
        {
            Version v = Assembly.GetExecutingAssembly().GetName().Version;
            string about = string.Format(CultureInfo.InvariantCulture, @"CommandLine Send Mail {0}.{1}.{2}  http://www.Stum.de", v.Major, v.Minor, v.Build);
            Console.WriteLine(about);
            Console.WriteLine(string.Empty.PadLeft(about.Length, ''));
        }

        /// <summary>
        /// Display the Help/Usage Text
        /// </summary>
        static void DisplayHelp()
        {
            // ReSharper disable PossibleNullReferenceException
            Console.WriteLine(string.Format(CultureInfo.InvariantCulture, "Usage: {0} <xmlfile>", Path.GetFileName(Process.GetCurrentProcess().MainModule.FileName)));
            // ReSharper restore PossibleNullReferenceException
        }

        /// <summary>
        /// Fetch the content of a given Node from the _settingsXml
        /// </summary>
        /// <param name="xpath">The XPath for this node</param>
        /// <returns>A String, containing the content of the Node or null if the node was not found</returns>
        static string FetchNodeData(string xpath)
        {
            var result = _settingsXml.SelectSingleNode(xpath);
            return result != null ? result.InnerText : null;
        }

        /// <summary>
        /// Fetches an Attribute for a given Xml Node
        /// </summary>
        /// <param name="xpath">The XPath to the node</param>
        /// <param name="attribute">The Name of the attribute</param>
        /// <returns>The value of the Attribute, or null if the attribute was not found</returns>
        static string FetchNodeAttribute(string xpath, string attribute)
        {
            var result = _settingsXml.SelectSingleNode(xpath);
            if (result != null && result.Attributes[attribute] != null)
            {
                return result.Attributes[attribute].InnerText;
            }
            return null;
        }

        /// <summary>
        /// Retreives all strings from a given tree
        /// </summary>
        /// <param name="xpath">The xpath to the tree</param>
        /// <returns>A List of strings from the tree</returns>
        static List<string> FetchNodesData(string xpath)
        {
            var result = new List<string>();
            XmlNodeList nodes = _settingsXml.SelectNodes(xpath);
            if (nodes != null)
            {
                foreach (XmlNode node in nodes)
                {
                    result.Add(node.InnerText);
                }
            }
            return result;
        }

        /// <summary>
        /// Get the desired Priority for this mail
        /// </summary>
        /// <returns>A valid MailPriority that can be used in a MailMessage</returns>
        static MailPriority GetPriority()
        {
            switch (FetchNodeData(@"/mail/priority"))
            {
                case "0":
                    return MailPriority.Low;
                case "2":
                    return MailPriority.High;
                default:
                    return MailPriority.Normal;
            }
        }

        /// <summary>
        /// Builds the MailMessage from the _settingsXml
        /// </summary>
        /// <remarks>
        /// We don't do error checking in the new MailAddress() cases, as
        /// invalid addresses will throw an exception that is being caught in
        /// Main() already.
        /// </remarks>
        /// <returns>A MailMessage that contains the data from the _settingsXml</returns>
        static MailMessage BuildMessage()
        {
            var result = new MailMessage();
            foreach (string address in FetchNodesData(@"/mail/to/*"))
            {
                if (!string.IsNullOrEmpty(address))
                    result.To.Add(new MailAddress(address));
            }
            foreach (string address in FetchNodesData(@"/mail/cc/*"))
            {
                if (!string.IsNullOrEmpty(address))
                    result.CC.Add(new MailAddress(address));
            }
            foreach (string address in FetchNodesData(@"/mail/bcc/*"))
            {
                if (!string.IsNullOrEmpty(address))
                    result.Bcc.Add(new MailAddress(address));
            }

            if (result.To.Count == 0 && result.CC.Count == 0 && result.Bcc.Count == 0)
            {
                throw new ArgumentException("No Receiver Addresses specified!");
            }

            foreach (string attachment in FetchNodesData(@"/mail/attachments/*"))
            {
                if (File.Exists(attachment))
                {
                    try
                    {
                        result.Attachments.Add(new Attachment(attachment));
                    }
                    catch (ArgumentException ex)
                    {
                        Console.Error.WriteLine(String.Format(CultureInfo.InvariantCulture, "ERROR: Attachment {0} is invalid! Message: {1}", attachment, ex.Message));
                    }
                }
            }

            string from = FetchNodeData(@"/mail/from");
            if (from == null)
            {
                throw new ArgumentException("No From Address specified!");
            }
            var timeStampSnapshot = DateTime.Now;
            bool replacebodytokens, replacesubjecttokens;
            if (!bool.TryParse(FetchNodeAttribute(@"/mail/body", "replacetokens"), out replacebodytokens))
            {
                replacebodytokens = false;
            }

            if (!bool.TryParse(FetchNodeAttribute(@"/mail/subject", "replacetokens"), out replacesubjecttokens))
            {
                replacesubjecttokens = false;
            }


            string replyto = FetchNodeData(@"/mail/replyto");
            bool isBodyHtml;
            if (bool.TryParse(FetchNodeAttribute(@"/mail/body", "html"), out isBodyHtml))
            {
                result.IsBodyHtml = isBodyHtml;
            }

            result.From = new MailAddress(from);

            var bodyBuilder = new StringBuilder();
            if (replacebodytokens)
            {
                bodyBuilder.Append(ReplaceTokens(FetchNodeData(@"/mail/body"), timeStampSnapshot));
            }
            else
            {
                bodyBuilder.Append(FetchNodeData(@"/mail/body"));
            }

            foreach (string embeddedfile in FetchNodesData(@"/mail/embedfiles/*"))
            {
                string embedcontent = ReadFile(embeddedfile);
                string embedfilename = Path.GetFileName(embeddedfile);
                if (isBodyHtml)
                {
                    bodyBuilder.AppendFormat("<br/><br/><u>{0}</u>", embedfilename);
                }
                else
                {
                    bodyBuilder.AppendFormat("\r\n\r\n{0}\r\n{1}\r\n", embedfilename, new String('-', embedfilename.Length));
                }

                if (isBodyHtml) bodyBuilder.Append("<pre>");
                if (embedcontent == null)
                {
                    bodyBuilder.Append("File was inaccessible or did not exist.");
                }
                else
                {
                    bodyBuilder.Append(embedcontent);
                }
                if (isBodyHtml) bodyBuilder.Append("</pre>");
            }
            
            result.Body = bodyBuilder.ToString();
            result.Subject = replacesubjecttokens
                                 ?
                                     ReplaceTokens(FetchNodeData(@"/mail/subject"), timeStampSnapshot)
                                 :
                                     FetchNodeData(@"/mail/subject");
            result.BodyEncoding = Encoding.UTF8;
            result.SubjectEncoding = Encoding.UTF8;
            result.Priority = GetPriority();
            result.ReplyTo = !string.IsNullOrEmpty(replyto) ? new MailAddress(replyto) : result.From;
            return result;
        }

        /// <summary>
        /// This replaces the Tokens in the Input String.
        /// </summary>
        /// <param name="input"></param>
        /// <param name="timeStampSnapshot">The DateTime to use when replacing date or time snapshots</param>
        /// <returns></returns>
        // ReSharper disable SuggestBaseTypeForParameter
        static string ReplaceTokens(string input,DateTime timeStampSnapshot)
        {
            string output = input;
            // ##datenow##
             output = output.Replace("##datenow##", timeStampSnapshot.ToString("yyyy-MM-dd",CultureInfo.InvariantCulture));
            // ##timenow##
             output = output.Replace("##timenow##", timeStampSnapshot.ToString("HH:mm:ss", CultureInfo.InvariantCulture));
            return output;
        }
        // ReSharper restore SuggestBaseTypeForParameter

        /// <summary>
        /// Builds the SmtpClient from the _settingsXml
        /// </summary>
        /// <returns>A SmtpClient</returns>
        static SmtpClient BuildSmtpClient()
        {
            var result = new SmtpClient();

            string portString = FetchNodeData(@"/mail/host/port");
            int port;
            if (!string.IsNullOrEmpty(portString) && int.TryParse(portString, out port))
            {
                result.Port = port;
            }

            result.Host = FetchNodeData(@"/mail/host/address");

            string username = FetchNodeData(@"/mail/host/username");
            string password = FetchNodeData(@"/mail/host/password");
            if (!string.IsNullOrEmpty(username) && !string.IsNullOrEmpty(password))
            {
                var authCredential = new NetworkCredential(username, password);
                result.UseDefaultCredentials = false;
                result.Credentials = authCredential;
            }
            return result;
        }

        /// <summary>
        /// Outputs Verbose Information about the mail
        /// </summary>
        /// <param name="client">The SmtpClient that is being sed</param>
        /// <param name="message">The MailMessage that is being sent</param>
        static void DisplayVerboseInfo(SmtpClient client, MailMessage message)
        {
            var attsb = new StringBuilder();
            for (int ai = 0; ai < message.Attachments.Count; ai++)
            {
                attsb.AppendFormat(CultureInfo.InvariantCulture, "{0}{1}", message.Attachments[ai].Name, ai == message.Attachments.Count - 1 ? "" : ", ");
            }

            var sb = new StringBuilder();
            sb.AppendFormat(CultureInfo.InvariantCulture, "Hostname: {0} (Using Auth: {1}){2}", client.Host, client.Credentials == null ? "no" : "yes", Environment.NewLine);
            sb.AppendFormat(CultureInfo.InvariantCulture, "From: {0} (Reply To: {1}){2}", message.From, message.ReplyTo, Environment.NewLine);
            sb.AppendFormat(CultureInfo.InvariantCulture, "To: {0}{1}", message.To, Environment.NewLine);
            sb.AppendFormat(CultureInfo.InvariantCulture, "Cc: {0}{1}", message.CC, Environment.NewLine);
            sb.AppendFormat(CultureInfo.InvariantCulture, "Bcc: {0}{1}", message.Bcc, Environment.NewLine);
            sb.AppendFormat(CultureInfo.InvariantCulture, "Attachments: {0}{1}", attsb, Environment.NewLine);
            sb.AppendFormat(CultureInfo.InvariantCulture, "Subject: {0}{1}", message.Subject, Environment.NewLine);
            sb.AppendFormat(CultureInfo.InvariantCulture, "Body (Html: {0}):{2}{1}", message.IsBodyHtml ? "yes" : "no", message.Body, Environment.NewLine);
            sb.Append(Environment.NewLine);
            Console.WriteLine(sb.ToString());
        }

        /// <summary>
        /// Read a file into a string
        /// </summary>
        /// <param name="filename">the file to load</param>
        /// <returns>Either the content of the file, or null if the file could not be opened.</returns>
        static string ReadFile(string filename)
        {
            string result = null;

            if (File.Exists(filename))
            {
                try
                {
                    using (var fs = File.OpenText(filename))
                    {
                        result = fs.ReadToEnd();
                        fs.Close();
                    }
                }
                catch (UnauthorizedAccessException ex)
                {
                    Console.Error.WriteLine(String.Format(CultureInfo.InvariantCulture, "ERROR: Embedded File {0} is invalid! Message: {1}", filename, ex.Message));
                    result = null;
                }
            }

            return result;
        }

        /// <summary>
        /// Main Function: Read and Parse Settings.xml and send mail
        /// </summary>
        /// <param name="args">The Command Line Parameters - needs to be exactly 1 Parameter: The Xml File with the Settings</param>
        /// <returns>An Exit Code as specified in the ExitCodes enum</returns>
        static int Main(string[] args)
        {
            DisplayAbout();
            if (args.Length != 1)
            {
                DisplayHelp();
                return (byte)ExitCode.InvalidArguments;
            }
            if (!File.Exists(args[0]))
            {
                Console.Error.WriteLine(string.Format(CultureInfo.InvariantCulture, "ERROR: File {0} does not exist!", args[0]));
                return (byte)ExitCode.SettingsXmlNotFound;
            }
            try
            {
                _settingsXml = new XmlDocument();
                _settingsXml.Load(args[0]);
                if (!bool.TryParse(FetchNodeAttribute(@"/mail", "verbose"), out _verbose))
                {
                    _verbose = false;
                }
                MailMessage message = BuildMessage();
                SmtpClient client = BuildSmtpClient();
                if (_verbose) DisplayVerboseInfo(client, message);
                Console.WriteLine("Trying to send Mail...");
                client.Send(message);
                return (byte)ExitCode.AllOk;
            }
            catch (XmlException ex)
            {
                Console.Error.WriteLine(string.Format(CultureInfo.InvariantCulture, "{2}ERROR: File {0} is invalid!{2}Message: {1}", args[0], ex.Message, Environment.NewLine));
                return (byte)ExitCode.SettingsXmlInvalid;
            }
            catch (SmtpFailedRecipientsException ex)
            {
                string message = ex.Message;
                if (ex.InnerException != null)
                {
                    message = ex.InnerException.Message;
                }
                Console.Error.WriteLine(string.Format(CultureInfo.InvariantCulture, "{2}ERROR: Could not send mail to {0}!{2}Message: {1}", ex.FailedRecipient, message, Environment.NewLine));
                return (byte)ExitCode.MailNotSentError;
            }
            catch (SmtpException ex)
            {
                string message = ex.Message;
                if (ex.InnerException != null)
                {
                    message = ex.InnerException.Message;
                }
                Console.Error.WriteLine(string.Format(CultureInfo.InvariantCulture, "{2}ERROR: There is a problem with the server {0}!{2}Message: {1}", FetchNodeData(@"/mail/host/address"), message, Environment.NewLine));
                return (byte)ExitCode.SmtpError;
            }
            catch (Exception ex)
            {
                Console.Error.WriteLine(string.Format(CultureInfo.InvariantCulture, "{1}ERROR: Unknown Error!{1}Message: {0}", ex.Message, Environment.NewLine));
                return (byte)ExitCode.UnknownError;
            }
            finally
            {
                Console.WriteLine(string.Format(CultureInfo.InvariantCulture, "{1}Finished execution on {0}", DateTime.Now.ToString("yyyy-MM-dd HH:mm:ss", CultureInfo.InvariantCulture), Environment.NewLine));
            }
        }
    }
}
